/*
 *      NOMAJORS...Expected percentage of acceptable lots in single,
 *      double, or triple stage TRINOMIAL sampling in which acceptance
 *      implies that NO MAJOR nonconformances were observed.
 *
 *               Bob Obenchain, CompuServe User [72007,467]
 *
 *      Parameter settings are either...
 *             specified interactively in response to prompts, or
 *             redirected to come from a batch input file.
 */

#include <stdio.h>
#include <string.h>
#include <process.h>
#include <math.h>

#define min(a,b)    (( (a) <= (b) ) ?  (a) : (b) )

#define NCONFVAL 99      /* max number of incoming % nonconforming values */

int            stages, samsiz1, accnum1, reject1;
int            samsiz2, accnum2, reject2, samsiz3, accnum3;
double         pa, pb, q, lp, lq;
double         split[5] = { 0.00, 0.10, 0.25, 0.50, 1.00 };
double         testeff, totnc, majnc, minnc, nctlast;
double         accprob[5][NCONFVAL], totalnc[NCONFVAL];
char           buf[256], inpnam[13], outnam[13];
FILE           *inpfile, *csvfile;

/* prototypes */

void   main( void );
double logcomb( int n, int k );
double cumprob( int accnum, int samsiz );
double sglprob( int accnum, int samsiz );
double dublsam( int newacc, int newsam, int accept, int reject, int samsiz );
double tripsam( void );

/* Logarithm of the Combinatorial Coefficient */

/*
 * NOTE: The Combinatorial Coefficient for choosing k objects from
 *       n is identical to the Trinomial Coefficient for Choosing ZERO
 *       and k objects from n...i.e. there's only 1 way to pick NONE.
 */

double logcomb( n, k )
int n, k; {

        int i;
        double  dk, dn, index, logc, log();

        logc = 0;
        if( k >= n || k <= 0 ) return( logc );

        if( n - k < k ) k = n - k;

        dk = (double)k;
        dn = (double)n;

        for( i=0; i<k; i++ ) {
                index = (double)i;
                logc += log( dn - index ) - log( dk - index );
                }

        return( logc );

        }  /* end logcomb */


/* Acceptance Probability Calculation for Single Stage Sampling */

double cumprob( accnum, samsiz )
int accnum, samsiz; {
        int    index;
        double sprob, cprob;
        double exp(), logcomb();

        /*
         * CUMPROB will be set equal to the probability that a TRINOMIAL
         * random variable with Major Nonconformance probability pa and
         * Minor Nonconformance probability pb in each of samsiz
         * independent trials will yield Zero Major and at most accnum Minor
         * nonconformances, where -
         * pa, pb, q=1-pa-pb, lp=log(pb), & lq=log(q) are in common storage.
         */

        cprob = 0;

        if( accnum < 0 || samsiz < 0 || accnum > samsiz ) return( cprob );

        for( index=0; index<=accnum; index++) {

                sprob = exp( logcomb( samsiz, index ) +
                             (double)index * lp +
                             (double)(samsiz-index) * lq );

                if( sprob > 1 )
                        sprob = 1;
                else if( sprob < 0 )
                        sprob = 0;

                cprob += sprob;
                }

        if( cprob > 1 )
               cprob = 1;
        else if( cprob < 0 )
               cprob = 0;

        /* printf( "\nCumprob for %6d, %6d = %6.4lf", accnum, samsiz, cprob); */

        return( cprob );

        }  /* end cumprob */


double sglprob( accnum, samsiz )
int accnum, samsiz; {
        double sprob;
        double exp(), logcomb();

        /*
         * SGLPROB will be set equal to the probability that a TRINOMIAL
         * random variable with Major Nonconformance probability pa and
         * Minor Nonconformance probability pb in each of samsiz
         * independent trials will yield Zero Major and exactly accnum Minor
         * nonconformances, where -
         * pa, pb, q=1-pa-pb, lp=log(pb), & lq=log(q) are in common storage.
         */
        sprob = 0;

        if( accnum < 0 || accnum > samsiz ) return( sprob );

        sprob = exp( logcomb( samsiz, accnum ) +
                     (double)accnum * lp +
                     (double)(samsiz-accnum) * lq );

        if( sprob > 1 )
                sprob = 1;
        else if( sprob < 0 )
                sprob = 0;

        /* printf( "\nSglprob for %6d, %6d = %6.4lf", accnum, samsiz, sprob); */

        return( sprob );

        }  /* end sglprob */


/* Acceptance Probability Calculation for Two Stage Sampling */

double dublsam( newacc, newsam, accept, reject, samsiz )
int newacc, newsam, accept, reject, samsiz; {
        int    index;
        double prob1, prob2, probacc, sglprob(), cumprob();

        probacc = cumprob( accept, samsiz );

        for( index=accept+1; index<reject; index++) {
                prob2 = cumprob( newacc-index, newsam );
                prob1 = sglprob( index, samsiz );
                probacc += prob1*prob2;
                }

        if( probacc > 1 )
                 probacc = 1;
        else if( probacc < 0 )
                 probacc = 0;

        /* printf( "\nDublsam for %6d, %6d, %6d, %6d, %6d = %6.4lf",
                newacc, newsam, accept, reject, samsiz, probacc); */

        return( probacc );

        }  /* end dublsam */


/* Acceptance Probability Calculation for Three Stage Sampling */

double tripsam() {
        int index;
        double prob1, prob23, probacc, sglprob(), dublsam();

        probacc = cumprob( accnum1, samsiz1 );

        for( index=accnum1+1; index<reject1; index++) {
                prob23   = dublsam( accnum3-index, samsiz3,
                                    accnum2-index, reject2, samsiz2 );
                prob1    = sglprob( index, samsiz1 );
                probacc += prob1*prob23;
                }

        if( probacc > 1 ) probacc = 1;
        else if( probacc < 0 ) probacc = 0;

        return( probacc );

        }  /* end tripsam */

void main() {

        char   kb;
        int    index, sndx;
        double cumprob(), dublsam(), tripsam();
        extern double log();
        extern FILE *fopen();                   /* Stream I/O */
        extern int fclose();                    /* Stream Output */

        printf("\n\n*** One, Two, or Three Stage TRINOMIAL Acceptance ***");
        printf("\n    Sampling in which the Observance of a Single");
        printf("\n    Major Nonconformance implies Lot Rejection.\n");
        printf("\n             NoMajors.EXE...Version 9402\n");
        printf("\n         A Quality Assurance Training Tool:");
        printf("\n  Statistics Committee of the QA Section of the PMA.");
        printf("\n\n      Bob Obenchain, CompuServe User [72007,467]\n\n" );

        printf(      "\tWill Parameter Settings be Input via K = Keyboard ? ...or" );
        printf(    "\n\t                                     B = Batch File ?" );
        printf(  "\n\n\tPress the   K   or   B   key now --> " );
        while( 1 ) {
              gets( buf );
              sscanf( buf, "%c", &kb );
              if( kb == 'k' )
                  kb  = 'K';
              if( kb == 'b' )
                  kb  = 'B';
              if( kb == 'K' || kb == 'B' )
                  break;
              printf( "\r\tPress either K  or  B  ...Try Again --> " );
              }

        strcpy( outnam, "nomajors" );
        strcpy( inpnam, "nomajors" );

        if( kb == 'B' )
            printf( "\n\n\tBatch Input of Parameter Settings Selected..." );
        else
            printf( "\n\n\tKeyboard Input of Parameter Settings Selected..." );

        printf( "\n\n\tAt colon Prompts : ...simply press ENTER to get the [default]." );

        if( kb == 'B' ) {
            printf(  "\n\n\tSpecify filename of Batch Input File [%s] : ", inpnam );
            gets( buf );
            if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
                sscanf( buf, "%s", inpnam );
            inpnam[ min( 8, (int)strcspn( inpnam, ". " ) ) ] = '\0';
            if( inpnam[0] == '\0' )
                strcpy( inpnam, "nomajors" );
            strcpy( outnam, inpnam );
            strcat( inpnam, ".in" );
            printf( "\n\tThe Batch Input file is to be: %s\n", inpnam );
            if( ( inpfile = fopen( inpnam, "r" ) ) == NULL ) {
                    printf(  "\tCannot read Batch Input file: %s\n", inpnam );
                    printf(  "\t...using Keyboard Input from standard Infile, stdin.\n" );
                    kb = 'K';
                    }
            }

        if( kb == 'K' )
            inpfile = stdin;

        printf(  "\n\n\tSpecify filename to save primary NoMajors Output [%s] : ",
            outnam );
        fgets( buf, 80, inpfile );
        if( buf[0] != '\0' && buf[0] != '\r' && buf[0] != '\n' )
            sscanf( buf, "%s", outnam );
        outnam[ min( 8, (int)strcspn( outnam, ". " ) ) ] = '\0';
        if( outnam[0] == '\0' )
            strcpy( outnam, "nomajors" );
        strcat( outnam, ".csv" );
        printf(  "\n\tThe NoMajors primary Output file is to be: %s\n",
                outnam );
        if( ( csvfile = fopen( outnam, "w" ) ) == NULL ) {
            printf(  "\tCannot write to Output filename : %s\n", outnam );
            if( ( csvfile = fopen( "nomajors.csv", "w" ) ) == NULL )
                    printf(  "\tCannot write to Output filename : nomajors.csv\n" );
            else {
                    printf(  "\t...using default Outfile name : nomajors.csv\n" );
                    strcpy( outnam, "nomajors.csv" );
                    }
            }

        stages = 1;

        printf("\n\nEnter First Stage Sample Size: ");
        fgets(buf, 100, inpfile);
        sscanf(buf, "%d", &samsiz1);
        if( samsiz1 <= 0 ) {
                printf("\nFirst Stage Sample Size will be 50, the default.\n");
                samsiz1 = 50;
                }
        else if( samsiz1 == 1 ) {
                printf("\nFirst Stage Sample Size will be 2, the minimum allowed.\n");
                samsiz1 = 2;
                }
        else if( samsiz1 > 5000 ) {
                printf("\nFirst Stage Sample Size will be 5,000...which is the maximum allowed here.\n");
                samsiz1 = 5000;
                }

        printf("\nEnter First Stage MINOR Nonconformance Accept Number: ");
        fgets(buf, 100, inpfile);
        sscanf(buf, "%d", &accnum1);
        if( accnum1 < -1 ) {
                printf("\nStage 1 Accept Number will be 0, the default.\n");
                accnum1 = 0;
                }
        else if( accnum1 == -1 ) {
                printf("\nA Stage 1 Accept Number of -1 indicates that the lot cannot be accepted\n");
                printf("at this first stage even if zero nonconformances are observed.\n");
                }
        else if( accnum1 >= samsiz1 ) {
                accnum1 = samsiz1/4;
                printf("\nStage 1 Accept Number will be %d, which is about one-fourth the Sample Size of %d.\n", accnum1, samsiz1);
                }
        if( accnum1 > 999 ) {
                printf("\nAccept Numbers cannot exceed 999.\n");
                accnum1 = 999;
                }

        printf("\nEnter First Stage MINOR Nonconformance Reject Number: ");
        fgets(buf, 100, inpfile);
        sscanf(buf, "%d", &reject1);
        if( reject1 < 1 ) reject1 = 1;
        if( reject1 <= accnum1 + 1 ) {
                if( accnum1 < 0 ) {
                        printf("\nAccept Number reset to Zero for Single Stage Sampling.\n");
                        accnum1 = 0;
                        }
                reject1 = accnum1 + 1;
                if( reject1 > samsiz1 ) {
                        reject1 = samsiz1;
                        printf("\nStage 1 Reject Number will be %d, ...all sampled units nonconform.\n", reject1);
                        }
                else {
                        printf("\nA Reject Number of %d is the minimum allowed for\n", reject1);
                        printf("Single Stage Sampling with Accept Number of %d.\n", accnum1);
                        }
                }

        if( reject1 > accnum1 + 1 ) {   /* Second Stage Required */

                stages = 2;

                printf("\nEnter Second Stage Sample Size: ");
                fgets(buf, 100, inpfile);
                sscanf(buf, "%d", &samsiz2);
                if( samsiz2 <= 0 ) {
                        printf("\nSecond Stage Sample Size will be 50, the default.\n");
                        samsiz2 = 50;
                        }
                else if( samsiz2 > 5000 ) {
                        printf("\nSecond Stage Sample Size will be 5,000...which is the maximum allowed here.\n");
                        samsiz2 = 5000;
                        }

                printf("\nEnter Second Stage MINOR Nonconformance Accept Number: ");
                fgets(buf, 100, inpfile);
                sscanf(buf, "%d", &accnum2);
                if( accnum2 < accnum1 ) {
                        accnum2 = accnum1;
                        printf("\nSecond Stage Accept Number will be %d, which is the minimum allowed when\n", accnum2);
                        printf("the First Stage Accept Number is %d.\n", accnum1);
                        }
                if( accnum2 < reject1 - 1 ) {
                        printf("\nA Third Stage is required when the Second Stage Accept Number is");
                        printf("\ntwo or more units smaller than the First Stage Reject Number.");
                        }
                if( accnum2 >= samsiz1 + samsiz2 ) {
                        accnum2 = (samsiz1 + samsiz2)/4;
                        printf("\nSecond Stage Accept Number will be %d, which is about one-fourth of the", accnum2);
                        printf("\n1st and 2nd Stage Sample Size of %d.\n", samsiz1+samsiz2);
                        }
                if( accnum2 > 999 ) {
                        printf("\nAccept Numbers cannot exceed 999.\n");
                        accnum2 = 999;
                        }

                printf("\nEnter Second Stage MINOR Nonconformance Reject Number: ");
                fgets(buf, 100, inpfile);
                sscanf(buf, "%d", &reject2);
                if( reject2 <= accnum2 + 1 ) {
                        reject2 = accnum2 + 1;
                        if( reject2 > samsiz1 + samsiz2 ) {
                                reject1 = samsiz1 + samsiz2;
                                printf("\nStage 2 Reject Number will be %d, ...all sampled units nonconform.\n", reject2);
                                }
                        else {
                                printf("\nA Reject Number of %d is the minimum allowed for\n", reject2);
                                printf("Two Stage Sampling with Accept Number of %d.\n", accnum2);
                                }
                        }
                }

        if( reject2 > accnum2 + 1 ) {   /* Third Stage Required */

                stages = 3;

                printf("\nEnter Third Stage Sample Size: ");
                fgets(buf, 100, inpfile);
                sscanf(buf, "%d", &samsiz3);
                if( samsiz3 <= 0 ) {
                        printf("\nThird Stage Sample Size will be 50, the default.\n");
                        samsiz3 = 50;
                        }
                else if( samsiz3 > 5000 ) {
                        printf("\nThird Stage Sample Size will be 5,000...which is the maximum allowed here.\n");
                        samsiz3 = 5000;
                        }

                printf("\nEnter Third Stage MINOR Nonconformance Accept Number: ");
                fgets(buf, 100, inpfile);
                sscanf(buf, "%d", &accnum3);
                if( accnum3 < reject2 - 1 ) {
                        accnum3 = reject2 - 1;
                        printf("\nThe Third Stage Accept Number will be %d, which is the minimum allowed when\n", accnum3);
                        printf("the Second Stage Reject Number is %d.\n", reject2);
                        }
                if( accnum3 >= samsiz1 + samsiz2 + samsiz3 ) {
                        accnum2 = (samsiz1 + samsiz2 + samsiz3)/4;
                        printf("\nThird Stage Accept Number will be %d, which is about one-fourth of the", accnum3);
                        printf("\nThree Stage Sample Total Size of %d.\n", samsiz1+samsiz2+samsiz3);
                        }
                if( accnum3 > 999 ) {
                        printf("\nAccept Numbers cannot exceed 999.\n");
                        accnum3 = 999;
                        }
                }

        printf("\nSpecify Testing Efficiency (Percentage): ");
        fgets(buf, 100, inpfile);
        sscanf(buf, "%lf", &testeff);
        testeff /= 100;
        if( testeff > 1 || testeff < 0.01 ) {
                printf("\nTesting Efficiency will be 100%%, the default.\n");
                testeff = 1;
                }
        else if( testeff < 0.1 ) {
                printf("\nTesting Efficiency will be 10%%, the minimum allowed here.\n");
                testeff = 0.1;
                }

        if( stages == 1 )
                printf("\n\nSUMMARY: One Stage Sampling...\n");
        else if( stages == 2 )
                printf("\n\nSUMMARY: Two Stage Sampling...\n");
        else
                printf("\n\nSUMMARY: Three Stage Sampling...\n");

        printf("\nFirst Stage Sample Size = %d", samsiz1 );
        fprintf( csvfile, "%4d, = First Stage Sample Size,,,,", samsiz1 );
        printf("\nMAJOR Nonconformance Reject Number = 1 at ALL Stages.");
        fprintf( csvfile, "\nMAJOR, Nonconformance Reject Number = 1, at ALL Stages.,,,");
        printf("\nFirst Stage MINOR Nonconformance Accept Number = %d", accnum1 );
        fprintf( csvfile, "\n%4d, = First Stage MINOR Nonconformance Accept Number,,,,", accnum1 );
        printf("\nFirst Stage MINOR Nonconformance Reject Number = %d", reject1 );
        fprintf( csvfile, "\n%4d, = First Stage MINOR Nonconformance Reject Number,,,,", reject1 );

        if( stages >= 2 ) {
                printf("\nSecond Stage Sample Size = %d", samsiz2 );
                fprintf( csvfile,"\n%4d, = Second Stage Sample Size,,,,", samsiz2 );
                printf("\nSecond Stage MINOR Nonconformance Accept Number = %d", accnum2 );
                fprintf( csvfile,"\n%4d, = Second Stage MINOR Nonconformance Accept Number,,,,", accnum2 );
                printf("\nSecond Stage MINOR Nonconformance Reject Number = %d", reject2 );
                fprintf( csvfile,"\n%4d, = Second Stage MINOR Nonconformance Reject Number,,,,", reject2 );
                }

        if( stages == 3 ) {
                printf("\nThird Stage Sample Size = %d", samsiz3 );
                fprintf( csvfile,"\n%4d, = Third Stage Sample Size,,,,", samsiz3 );
                printf("\nFinal MINOR Nonconformance Accept Number = %d", accnum3 );
                fprintf( csvfile,"\n%4d, = Final MINOR Nonconformance Accept Number,,,,", accnum3 );
                }

        printf("\nTesting Efficiency = %6.1lf%%", 100*testeff );
        fprintf( csvfile, "\n%6.1lf, = Testing Efficiency %%,,,,", 100*testeff );

        printf("\n\nPress a KEY to start calculations...");
        fprintf( csvfile, "\n         ,%%Yield@ ,%%Yield@ ,%%Yield@ ,%%Yield@ ,%%Yield@");
        fprintf( csvfile, "\nIn%%TOTAL,00%%Major,10%%Major,25%%Major,50%%Major,100%%Major");
        fprintf( csvfile, "\n   0.00,100.00,100.00,100.00,100.00,100.00" );
        getchar();

        nctlast = 0.0;
        totnc = 0.1;

        for( index=0; index<NCONFVAL; index++ ) {

            printf( "\n\nStep Number = %d", index );
            
            printf( "\n\nEnter TOTAL (Major+Minor) Nonconformance Percentage: " );
            fgets(buf, 100, inpfile);
            sscanf(buf, "%lf", &totnc);
            
            if( totnc < 0 )
                break;
            else if( totnc >= 100 ) {
                printf("\nTotal Nonconformance Percentage has been set to 99%%.\n");
                totnc = 99;
                }
            if( totnc == nctlast )
                index--;
            else
                totalnc[index] = totnc;
            
            for( sndx = 0; sndx < 5; sndx++ ) {

                majnc = split[sndx] * totnc;
                minnc = totnc - majnc;
                pa = testeff*majnc/100;  /* convert % to fraction */
                pb = testeff*minnc/100;  /* convert % to fraction */
            
                if( totnc != nctlast ) {
                    /*
                     * perform calculations only when total incoming
                     * nonconformance percentage has changed from
                     * last step.
                     */
                    if( pa == 0 && pb == 0 )
                        accprob[sndx][index] = 1.0;
                    else if( pa + pb >= 1 )
                        accprob[sndx][index] = 0.0;
                    else {

                        if( pb < 0.0000000001 )
                            pb = 0.0000000001;

                        /* to handle the cases where pa > 0 but pb zero. */

                        lp = log( pb );
                        q = 1 - pa - pb;
                        lq = log( q );

                        if( stages == 1 )
                                accprob[sndx][index] =
                                    cumprob( accnum1, samsiz1 );
                        else if( stages == 2 )
                                accprob[sndx][index] =
                                    dublsam( accnum2, samsiz2,
                                             accnum1, reject1, samsiz1 );
                        else if( stages == 3 )
                                accprob[sndx][index] = tripsam();
                        }
                    }

                printf( "\n\n For Incoming TOTAL Nonconformance Percentage = %6.2lf,",
                        totnc );
                printf( "\n%3.0lf%% Incoming MAJOR Nonconformance Percentage = %6.2lf,",
                        100.0*split[sndx], majnc );
                printf( "\n and Incoming MINOR Nonconformance Percentage = %6.2lf,",
                        minnc );
                printf( "\n Expected First-Pass Yield Percentage   = %6.2lf.\n",
                        100*accprob[sndx][index] );
                if( inpfile == stdin ) {
                    printf( "\n  Press ENTER..." );
                    gets( buf);
                    }
                }

            if( totnc != nctlast ) {
                /* Successive duplicates not writtem to output file. */
                fprintf( csvfile, "\n %6.2lf,%6.2lf,%6.2lf,%6.2lf,%6.2lf,%6.2lf",
                    totnc, 100*accprob[0][index], 100*accprob[1][index],
                    100*accprob[2][index], 100*accprob[3][index],
                    100*accprob[4][index] );
                }
            nctlast = totnc;
            
            printf( "\n\nTo QUIT, enter a NEGATIVE Total Nonconformance Percentage..." );
            }
            
        fprintf( csvfile, "\n -1, ...End of Input Signal...,,,," );
        fclose( csvfile );
        exit(0);
        }
